from typing import Union, List, Tuple, Sequence

from .loader import Loader
from .dialect import Dialect
from .constant import LIMIT_1


def get_table_select_sql(table_name: str, where: str, order_by: str, limit: Union[int, Tuple[int], List[int]], *columns):
	columns = Dialect.get_dialect_str(columns) if columns else Dialect.get_table_columns(table_name)
	table_name = Dialect.get_dialect_str(table_name)
	if where:
		sql = f'SELECT {columns} FROM {table_name} {where}'
	else:
		sql = f'SELECT {columns} FROM {table_name}'
	
	if order_by:
		sql = f'{sql} ORDER BY {order_by}'
	
	if limit:
		if isinstance(limit, int):
			return f'{sql} LIMIT ?'
		elif isinstance(limit, (Tuple, List)) and len(limit) == 2:
			return Dialect.limit_offset_sql(sql)
		else:
			raise ValueError("The type of the parameter 'limit' must be 'int' or tuple, list, and it length is 2.")
	
	return sql


def get_condition_arg(k: str, v: object):
	if k.endswith("__eq"):
		return f"{Dialect.get_dialect_str(k[:-4])} = ?", v
	if k.endswith("__ne"):
		return f"{Dialect.get_dialect_str(k[:-4])} != ?", v
	if k.endswith("__gt"):
		return f"{Dialect.get_dialect_str(k[:-4])} > ?", v
	if k.endswith("__lt"):
		return f"{Dialect.get_dialect_str(k[:-4])} < ?", v
	if k.endswith("__ge"):
		return f"{Dialect.get_dialect_str(k[:-4])} >= ?", v
	if k.endswith("__gte"):
		return f"{Dialect.get_dialect_str(k[:-5])} >= ?", v
	if k.endswith("__le"):
		return f"{Dialect.get_dialect_str(k[:-4])} <= ?", v
	if k.endswith("__lte"):
		return f"{Dialect.get_dialect_str(k[:-5])} <= ?", v
	if k.endswith("__isnull"):
		return f"{Dialect.get_dialect_str(k[:-8])} is {'null' if v else 'not null'}", None
	if k.endswith("__in") and isinstance(v, Sequence) and not isinstance(v, str):
		return f"{Dialect.get_dialect_str(k[:-4])} in({','.join(['?' for _ in v])})", v
	if k.endswith("__in"):
		return f"{Dialect.get_dialect_str(k[:-4])} in(?)", v
	if k.endswith("__not_in") and isinstance(v, Sequence) and not isinstance(v, str):
		return f"{Dialect.get_dialect_str(k[:-8])} not in({','.join(['?' for _ in v])})", v
	if k.endswith("__not_in"):
		return f"{Dialect.get_dialect_str(k[:-8])} not in(?)", v
	if k.endswith("__like"):
		return f"{Dialect.get_dialect_str(k[:-6])} like ?", f'%{v}%'
	if k.endswith("__startswith"):
		return f"{Dialect.get_dialect_str(k[:-12])} like ?", f'{v}%'
	if k.endswith("__endswith"):
		return f"{Dialect.get_dialect_str(k[:-10])} like ?", f'%{v}'
	if k.endswith("__contains"):
		return f"{Dialect.get_dialect_str(k[:-10])} like ?", f'%{v}%'
	if k.endswith("__range") and isinstance(v, Sequence) and 2 == len(v) and not isinstance(v, str):
		col = k[:-7]
		col_name = Dialect.get_dialect_str(col)
		return f"{col_name} >= ? and {col_name} <= ?", v
	if k.endswith("__between") and isinstance(v, Sequence) and 2 == len(v) and not isinstance(v, str):
		return f"{Dialect.get_dialect_str(k[:-9])} between ? and ?", v
	if k.endswith("__range") or k.endswith("__between"):
		return ValueError("Must is instance of Sequence with length 2 when use range or between statement")
	
	return f"{Dialect.get_dialect_str(k)} = ?", v


def get_where_arg_limit(**kwargs):
	where, args, limit = '', [], 0
	if 'limit' in kwargs:
		limit = kwargs.pop('limit')
	
	if kwargs:
		conditions, tmp_args = zip(*[get_condition_arg(k, v) for k, v in kwargs.items()])
		tmp_args = [arg for arg in tmp_args if arg is not None]
		
		for arg in tmp_args:
			if arg is not None:
				if isinstance(arg, Sequence) and not isinstance(arg, str):
					args.extend(arg)
				else:
					args.append(arg)
		where = f"WHERE {' and '.join(conditions)}"
	
	return where, args, limit


class WhereBase:

	def __init__(self, _exec, table_name: str, order_by: str, **kwargs):
		self.exec = _exec
		self.table = table_name
		self._order_by = order_by
		self.where_condition = kwargs
		self._limit = None
	
	def select(self, *columns) -> List[Tuple]:
		"""
		Select data from table and return list results(tuple).

		Examples
		--------
		>>> import sqlexecx as db
		>>> db.table('person').where(name='张三', age=20).select('id', 'name', 'age')
		[(3, '张三', 20)]
		"""
		sql, args = self.get_select_sql_args(*columns)
		return self.exec.do_select(sql, *args)
	
	def query(self, *columns) -> List[dict]:
		"""
		Select data from table and return list results(tuple).

		Examples
		--------
		>>> import sqlexecx as db
		>>> db.table('person').where(name='张三', age=20).query('id', 'name', 'age')
		[{'id': 3, 'name': '张三', 'age': 20}]
		"""
		sql, args = self.get_select_sql_args(*columns)
		return self.exec.do_query(sql, *args)

	def load(self, *columns) -> Loader:
		"""
		Get a Loader instance
		
		Examples
		--------
		>>> import sqlexecx as db
		>>> db.table('person').where(name='张三', age=20).load('id', 'name', 'age')
		Loader()
		"""
		sql, args = self.get_select_sql_args(*columns)
		return self.exec.do_load(sql, *args)
	
	def get_select_sql_args(self, *columns):
		where, args, limit = get_where_arg_limit(**self.where_condition)
		limit = self._limit or limit
		sql = get_table_select_sql(self.table, where, self._order_by, limit, *columns)
		if limit:
			if isinstance(limit, int):
				args = [*args, limit]
			else:
				args = [*args, *limit]
		return sql, args
	
	def get_select_one_sql_args(self, *columns):
		where, args, _ = get_where_arg_limit(**self.where_condition)
		sql = get_table_select_sql(self.table, where, self._order_by, LIMIT_1, *columns)
		return sql, args



from typing import Collection, Sequence, Union

t = ('1', '2')
print(isinstance(t, Collection), isinstance(t, Sequence))
# True True
l = ['1', '2']
print(isinstance(l, Collection), isinstance(l, Sequence))
# True True
s = set(l)
print(isinstance(s, Collection), isinstance(s, Sequence))
# True False
s = 'hello'
print(isinstance(s, Collection), isinstance(s, Sequence))
# True True
print('='*50)


def get_dialect_str(src: Union[str, Collection[str]]):
    return src.strip() if isinstance(src, str) else ','.join([s.strip() for s in src])


def _dialect_str(src: str):
    assert isinstance(src, str), 'source must be a string.'
    if ',' in src:
        return ','.join(['`{}`'.format(s.strip()) for s in src.split(',')])
    elif '(' in src:
        return src.strip()
    else:
        return '`{}`'.format(src.strip())


def get_mysql_dialect_str(src: Union[str, Collection[str]]):
    assert src, 'src string is required.'
    if isinstance(src, str):
        return _dialect_str(src)
    return ','.join([_dialect_str(arg) for arg in src])


if __name__ == '__main__':
    # names = 'person'
    # names = ['id', 'name, age']
    # print(get_dialect_str(names))
    # print(get_mysql_dialect_str(names))

    sql = """
    SELECT id, strftime('%Y-%m-%d %H:%M:%S' , create_time, 'localtime')
      FROM history where user_id = ? and query_type = ? and name like 'lisi%' and name like '%zhangsan' order by id desc
    """.strip()
    
    # import re
    # # NAMED_REGEX = r':[a-z|A-Z|\d][\w|\d]*'
    # NAMED_REGEX = r"like\s+'[\w|\d|%]*'"
    # pattern = re.compile(NAMED_REGEX, re.I)
    # for r in re.finditer(pattern, sql):
    #     old = r.group()
    #     new = old.replace('%', '%%').replace('%%%%', '%%')
    #     print(old, new)
    #     sql = sql.replace(old, new, 1)
        # print(r)
        # print(r.span())
        # print(r.group())
    # print(re.search(pattern, sql))
    # print(sql)

    # from sqlexecx import Engine, Dialect
    # Dialect.init(Engine.MYSQL)
    # print(Dialect.before_execute(sql))
    
    import sqlexecx as db
    db.init('/Users/summy/project/python/streamlit_clip_annoy/data/db.sqlite3', show_sql=True)
    print(db.table('history').columns("id, strftime('%Y-%m-%d %H:%M:%S', create_time, 'localtime') as create_time").query())
    


from typing import Collection, Iterable, List, Tuple, Mapping
from .loader import Loader
from .dialect import Dialect
from . import sql_support
from .constant import MODULE
from .log_support import logger, insert_log, save_log, batch_sql_log
from sqlexecutorx import execute as _execute, select as _select, select_one as _select_one, do_select as _do_select,\
    save as _save, batch_execute as _batch_execute, get as _get, query as _query, query_one as _query_one, DBError


def execute(sql: str, *args, **kwargs) -> int:
    """
    Execute sql return effect rowcount

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'INSERT INTO person(name, age) VALUES(?, ?)'
    >>> db.execute(sql, '张三', 20)
    1
    >>> sql = 'INSERT INTO person(name, age) VALUES(:name, :age)'
    >>> db.execute(sql, name='张三', age=20)
    1
    """

    sql, args = sql_support.try_mapping('sqlexecx.execute', sql, *args, **kwargs)
    return do_execute(sql, *args)


def insert(table_name: str, **kwargs) -> int:
    """
    Insert data into table, return effect rowcount.

    :param table_name: table name
    :param kwargs: {name='张三', age=20}
    return: Effect rowcount

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.insert('person', name='李四', age=18)
    1
    """

    insert_log('insert', table_name, **kwargs)
    sql, args = sql_support.insert_sql_args(table_name.strip(), **kwargs)
    return _execute(sql, *args)


def save(table_name: str, **kwargs):
    """
    Insert data into table, return primary key.

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.save('person', name='李四', age=18)
    4
    """

    try:
        select_key = Dialect.get_select_key(table_name=table_name)
    except NotImplementedError:
        raise DBError(f"Expect 'select_key' but not. you may should use 'save_select_key' func with 'select_key'.")
    return save_select_key(select_key, table_name, **kwargs)


def save_sql(sql: str, *args, **kwargs):
    """
    Insert data into table, return primary key.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'INSERT INTO person(name, age) VALUES(?, ?)'
    >>> db.save_sql(sql, '张三', 20)
    3
    >>> sql = 'INSERT INTO person(name, age) VALUES(:name, :age)'
    >>> db.save_sql(sql, name='张三', age=20)
    3
    """

    try:
        select_key = Dialect.get_select_key(sql=sql)
    except NotImplementedError:
        raise DBError(f"Expect 'select_key' but not. you may should use 'save_sql_select_key' func with 'select_key'.")
    return save_sql_select_key(select_key, sql, *args, **kwargs)


def save_select_key(select_key: str, table_name: str, **kwargs):
    """
    Insert data into table, return primary key.

    :param select_key: sql for select primary key
    :param table_name: table name
    :param kwargs: {name='张三', age=20}
    :return: Primary key

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.save_select_key('SELECT LAST_INSERT_ID()', 'person', name='李四', age=18)
    4
    """

    save_log('save_select_key', select_key, table_name, **kwargs)
    sql, args = sql_support.insert_sql_args(table_name.strip(), **kwargs)
    return save_sql_select_key(select_key, sql, *args)


def save_sql_select_key(select_key: str, sql: str, *args, **kwargs):
    """
    Insert data into table, return primary key.

    :param select_key: sql for select primary key
    :param sql: SQL
    :return: Primary key

    Examples
    --------
    >>> import sqlexecx as db
    >>> select_key = 'SELECT LAST_INSERT_ID()'
    >>> sql = 'INSERT INTO person(name, age) VALUES(?, ?)'
    >>> db.save_sql_select_key(select_key, sql, '张三', 20)
    3
    >>> sql = 'INSERT INTO person(name, age) VALUES(:name, :age)'
    >>> db.save_sql_select_key(select_key, sql, name='张三', age=20)
    3
    """

    logger.debug("Exec func 'sqlexecx.%s', 'select_key': %s \n\t sql: %s \n\t args: %s \n\t kwargs: %s" % ('save_sql', select_key, sql, args, kwargs))
    sql, args = sql_support.get_mapping_sql_args(sql, *args, **kwargs)
    return do_save_sql_select_key(select_key, sql, *args)


def get(sql: str, *args, **kwargs):
    """
    Execute select SQL and expected one int and only one int result, SQL contain 'limit'.

    MultiColumnsError: Expect only one column.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT count(1) FROM person WHERE name=? and age=? LIMIT 1'
    >>> db.get(sql, '张三', 20)
    1
    >>> sql = 'SELECT count(1) FROM person WHERE name=:name and age=:age LIMIT 1'
    >>> db.get(sql, name='张三', age=20)
    1
    """

    sql, args = sql_support.try_mapping('sqlexecx.get', sql, *args, **kwargs)
    return do_get(sql, *args)


def select(sql: str, *args, **kwargs) -> List[Tuple]:
    """
    Execute select SQL and return list results(tuple).

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.select(sql, '张三', 20)
    [(3, '张三', 20)]
    >>> sql = 'SELECT id, name, age FROM person WHERE name=:name and age=:age '
    >>> db.get(sql, name='张三', age=20)
    [(3, '张三', 20)]
    """

    sql, args = sql_support.try_mapping('sqlexecx.select', sql, *args, **kwargs)
    return do_select(sql, *args)


def select_one(sql: str, *args, **kwargs) -> Tuple:
    """
    Execute select SQL and return unique result(tuple), SQL contain 'limit'.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=? LIMIT 1'
    >>> db.select_one(sql, '张三', 20)
    (3, '张三', 20)
    >>> sql = 'SELECT id, name, age FROM person WHERE name=:name and age=:age LIMIT 1'
    >>> db.select_one(sql, name='张三', age=20)
    (3, '张三', 20)
    """

    sql, args = sql_support.try_mapping('sqlexecx.select_one', sql, *args, **kwargs)
    return do_select_one(sql, *args)


def query(sql: str, *args, **kwargs) -> List[dict]:
    """
    Execute select SQL and return list results(dict).

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.query(sql, '张三', 20)
    [{'id': 3, 'name': '张三', 'age': 20}]
    >>> sql = 'SELECT id, name, age FROM person WHERE name=:name and age=:age '
    >>> db.query(sql, name='张三', age=20)
    [{'id': 3, 'name': '张三', 'age': 20}]
    """
    sql, args = sql_support.try_mapping('sqlexecx.query', sql, *args, **kwargs)
    return do_query(sql, *args)


def query_one(sql: str, *args, **kwargs) -> dict:
    """
    Execute select SQL and return unique result(dict), SQL contain 'limit'.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=? LIMIT 1'
    >>> db.query_one(sql, '张三', 20)
    {'id': 3, 'name': '张三', 'age': 20}
    >>> sql = 'SELECT id, name, age FROM person WHERE name=:name and age=:age LIMIT 1'
    >>> db.query_one(sql, name='张三', age=20)
    {'id': 3, 'name': '张三', 'age': 20}
    """
    sql, args = sql_support.try_mapping('sqlexecx.query_one', sql, *args, **kwargs)
    return do_query_one(sql, *args)


def do_execute(sql: str, *args) -> int:
    """
    Execute sql return effect rowcount

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'INSERT INTO person(name, age) VALUES(?, ?)'
    >>> db.do_execute(sql, '张三', 20)
    1
    """
    sql = Dialect.before_execute(sql)
    return _execute(sql, *args)


def do_save_sql(sql: str, *args):
    """
    Insert data into table, return primary key.

    :param sql: SQL
    :param args:
    :return: Primary key

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'INSERT INTO person(name, age) VALUES(?, ?)'
    >>> db.do_save_sql(sql, '张三', 20)
    3
    """

    try:
        select_key = Dialect.get_select_key(sql=sql)
    except NotImplementedError:
        raise DBError(f"Expect 'select_key' but not. you may should use 'save_sql_select_key' func with 'select_key'.")
    return do_save_sql_select_key(select_key, sql, *args)


def do_save_sql_select_key(select_key: str, sql: str, *args):
    """
    Insert data into table, return primary key.

    :param select_key: sql for select primary key
    :param sql: SQL
    :param args:
    :return: Primary key

    Examples
    --------
    >>> import sqlexecx as db
    >>> select_key = 'SELECT LAST_INSERT_ID()'
    >>> sql = 'INSERT INTO person(name, age) VALUES(?, ?)'
    >>> db.do_save_sql_select_key(select_key, sql, '张三', 20)
    3
    """

    sql = Dialect.before_execute(sql)
    return _save(select_key, sql, *args)


def do_get(sql: str, *args):
    """
    Execute select SQL and expected one int and only one int result, SQL contain 'limit'.

    MultiColumnsError: Expect only one column.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT count(1) FROM person WHERE name=? and age=? LIMIT 1'
    >>> db.do_get(sql, '张三', 20)
    1
    """

    sql, args = sql_support.limit_one_sql_args(sql, *args)
    sql = Dialect.before_execute(sql)
    return _get(sql, *args)


def do_select(sql: str, *args) -> List[Tuple]:
    """
    Execute select SQL and return list results(tuple).

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.do_select(sql, '张三', 20)
    [(3, '张三', 20)]
    """

    sql = Dialect.before_execute(sql)
    return _select(sql, *args)


def do_select_one(sql: str, *args) -> Tuple:
    """
    Execute select SQL and return unique result(tuple), SQL contain 'limit'.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=? LIMIT 1'
    >>> db.do_select_one(sql, '张三', 20)
    (3, '张三', 20)
    """

    sql, args = sql_support.limit_one_sql_args(sql, *args)
    sql = Dialect.before_execute(sql)
    return _select_one(sql, *args)


def do_query(sql: str, *args) -> List[dict]:
    """
    Execute select SQL and return list results(dict).

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.do_query(sql, '张三', 20)
    [{'id': 3, 'name': '张三', 'age': 20}]
    """

    sql = Dialect.before_execute(sql)
    return _query(sql, *args)


def do_query_one(sql: str, *args) -> dict:
    """
    Execute select SQL and return unique result(dict), SQL contain 'limit'.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=? LIMIT 1'
    >>> db.do_query_one(sql, '张三', 20)
    {'id': 3, 'name': '张三', 'age': 20}
    """

    sql, args = sql_support.limit_one_sql_args(sql, *args)
    sql = Dialect.before_execute(sql)
    return _query_one(sql, *args)


def select_page(sql: str, page_num=1, page_size=10, *args, **kwargs) -> List[Tuple]:
    """
    Execute select SQL and return list(tuple) or empty list if no result.

    Automatically add 'limit ?,?' after sql statement if not.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.select_page(sql, 1, 10, '张三', 20)
    [(3, '张三', 20)]
    >>> sql = 'SELECT id, name, age FROM person WHERE name=:name and age=:age '
    >>> db.select_page(sql, 1, 10, name='张三', age=20)
    [(3, '张三', 20)]
    """

    sql, args = sql_support.try_mapping('select_page', sql, *args, **kwargs)
    return do_select_page(sql, page_num, page_size, *args)


def query_page(sql: str, page_num=1, page_size=10, *args, **kwargs) -> List[dict]:
    """
    Execute select SQL and return list or empty list if no result.

    Automatically add 'limit ?,?' after sql statement if not.

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.query_page(sql, 1, 10, '张三', 20)
    [{'id': 3, 'name': '张三', 'age': 20}]
    >>> sql = 'SELECT id, name, age FROM person WHERE name=:name and age=:age '
    >>> db.query_page(sql, 1, 10, name='张三', age=20)
    [{'id': 3, 'name': '张三', 'age': 20}]
    """
    sql, args = sql_support.try_mapping('query_page', sql, *args, **kwargs)
    return do_query_page(sql, page_num, page_size, *args)


def do_select_page(sql: str, page_num=1, page_size=10, *args) -> List[Tuple]:
    """
    Execute select SQL and return list results(dict).

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.do_select_page(sql, 1, 10, '张三', 20)
    [(3, '张三', 20)]
    """

    sql, args = Dialect.get_page_sql_args(sql, page_num, page_size, *args)
    return do_select(sql, *args)


def do_query_page(sql: str, page_num=1, page_size=10, *args) -> List[dict]:
    """
    Execute select SQL and return list results(dict).

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.do_query_page(sql, 1, 10, '张三', 20)
    [{'id': 3, 'name': '张三', 'age': 20}]
    """

    sql, args = Dialect.get_page_sql_args(sql, page_num, page_size, *args)
    return do_query(sql, *args)


def batch_execute(sql: str, *args) -> int:
    """
    Batch execute

    :param sql: SQL to execute
    :param args: All number must have same size.
    :return: Effect row count

    Examples
    --------
    >>> import sqlexecx as db
    >>> args = [('张三', 20), ('李四', 28)]
    >>> sql = 'INSERT INTO person(name, age) VALUES(?, ?)'
    >>> db.batch_execute(sql, *args)
    2
    >>> args =  [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
    >>> sql = 'INSERT INTO person(name, age) VALUES(:name, :age)'
    >>> db.batch_execute(sql, *args)
    2
    """

    batch_sql_log(MODULE, 'batch_execute', sql, args)
    assert args, "*args must not be empty."
    args = sql_support.get_batch_args(*args)
    if isinstance(args[0], Mapping):
        sql, args = sql_support.batch_named_sql_args(sql, *args)
    sql = Dialect.before_execute(sql)
    args = sql_support.get_batch_args(*args)
    return _batch_execute(sql, *args)


def batch_insert(table_name: str, *args) -> int:
    """
    Batch insert data into table and return effect rowcount

    :param table_name: table name
    :param args: All number must have same key. [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
    :return: Effect row count

    Examples
    --------
    >>> import sqlexecx as db
    >>> args =  [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
    >>> db.batch_execute('person', args)
    2
    >>> db.batch_execute('person', *args)
    2
    """
    
    logger.debug("Exec func 'sqlexecx.%s' \n\t Table: '%s', args: %s" % ('batch_insert', table_name, args))
    assert args, "*args must not be empty."
    args = sql_support.get_batch_args(*args)
    assert isinstance(args[0], Mapping), 'args must be a collection of Mapping'
    sql, args = sql_support.batch_insert_sql_args(table_name, *args)
    return _batch_execute(sql, *args)


def load(sql: str, *args, **kwargs) -> Loader:
    """
    Execute select SQL and return a Loader instance

    :param sql: SQL to be executed
    :return: Loader

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.load(sql, '张三', 20)
    Lodder()
    >>> sql = 'SELECT id, name, age FROM person WHERE name=:name and age=:age'
    >>> db.load(sql, name='张三', age=20)
    Lodder()
    """
    sql, args = sql_support.try_mapping('sqlexecx.csv', sql, *args, **kwargs)
    return do_load(sql, *args)


def do_load(sql: str, *args) -> Loader:
    """
    Execute select SQL and return a Loader instance

    :param sql: SQL to be executed
    :return: Loader

    Examples
    --------
    >>> import sqlexecx as db
    >>> sql = 'SELECT id, name, age FROM person WHERE name=? and age=?'
    >>> db.do_load(sql, '张三', 20)
    Lodder()
    """
    sql = Dialect.before_execute(sql)
    return Loader(*_do_select(sql, *args))


def insert_from_df(df, table_name: str, columns: Collection[str] = None) -> int:
    """
    Insert data into table from pandas DataFrame and return effected rowcount.

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.insert_from_df(df, 'person')
    """

    columns = columns if columns and len(columns) > 0 else df.columns.tolist()
    sql = sql_support.insert_sql(table_name.strip(), columns)
    return batch_execute(sql, df.values.tolist())


def insert_from_csv(file_name: str, table_name: str, delimiter=',', header=True, columns: Collection[str] = None, encoding='utf-8') -> int:
    """
    Insert data into table from a csv file and return effected rowcount.

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.insert_from_csv('test.csv', 'person')
    20
    """

    import csv
    sql = None
    if columns and len(columns) > 0:
        sql = sql_support.insert_sql(table_name.strip(), columns)
    elif not header:
        raise ValueError("Expected one of 'header' and 'columns'.")

    with open(file_name, newline='', encoding=encoding) as f:
        lines = csv.reader(f, delimiter=delimiter)
        lines = [line for line in lines]

    if len(lines) == 0:
        return 0

    if header:
        if len(lines) == 1:
            return 0

        if sql is None:
            sql = sql_support.insert_sql(table_name.strip(), lines[0])
        lines = lines[1:]

    return batch_execute(sql, lines)


def insert_from_json(file_name: str, table_name: str, encoding='utf-8') -> int:
    """
    Insert data into table from a json file and return effected rowcount.

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.insert_from_json('test.json', 'person')
    20

    many rows json file example:
    [{"id": 1, "name": "张三", "age": 55}, ...]

    one row json file example:
    {"id": 1, "name": "张三", "age": 55}
    """

    import json

    with open(file_name, encoding=encoding) as f:
        data = json.load(f)

    if isinstance(data, dict):
        return insert(table_name, **data)
    elif isinstance(data, Iterable):
        return batch_insert(table_name, data)
    else:
        logger.info("Exec func 'sqlexecx.%s' \n\t Table: '%s' insert 0 rows." % ('insert_from_json', table_name))
        return 0


def truncate_table(table_name: str) -> int:
    """
    Truncate table

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.truncate_table('person')
    """
    return Dialect.truncate_table(table_name)


def drop_table(table_name: str) -> int:
    """
    Drop table

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.drop_table('person')
    """
    return _execute('DROP TABLE IF EXISTS %s' % Dialect.get_dialect_str(table_name))


def show_tables(schema: str = None) -> List[str]:
    """
    Show tables

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.show_tables()
    """
    return Dialect.show_tables(schema)



from sqlexecutorx.constant import LIMIT_1, CACHE_SIZE

MODULE = 'SQLExecX'

SELECT_COUNT = 'count(1)'

NAMED_REGEX = r':[a-z|A-Z|\d][\w|\d]*'

DEFAULT_KEY_FIELD = 'id'

MYSQL_SELECT_KEY = "SELECT LAST_INSERT_ID()"

SQLITE_SELECT_KEY = 'SELECT last_insert_rowid()'

MYSQL_COLUMN_SQL = '''SELECT GROUP_CONCAT(CONCAT("`",column_name,"`") SEPARATOR ",") 
                        FROM `information_schema`.`columns` WHERE `table_schema` = (SELECT DATABASE()) AND `table_name` = %s LIMIT %s'''

POSTGRES_COLUMN_SQL = '''SELECT array_to_string(array_agg(column_name),',') as column_name FROM information_schema.columns 
                          WHERE table_schema='public' and table_name = %s LIMIT %s'''


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

from typing import List, Tuple
from functools import lru_cache

from . import exec
from .loader import Loader
from .page_exec import PageExec
from .constant import CACHE_SIZE


class SqlPageExec:

    def __init__(self, sql: str, page_exec: PageExec):
        self.sql = sql
        self.page_exec = page_exec

    def query(self, *args, **kwargs) -> List[dict]:
        """
        Execute select SQL and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').page(1, 10).query('张三', 20)
        [{'id': 3, 'name': '张三', 'age': 20}]
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').page(1, 10).query(name='张三', age=20)
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        return self.page_exec.query(self.sql, *args, **kwargs)

    def select(self, *args, **kwargs) -> List[Tuple]:
        """
        Execute select SQL and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').page(1, 10).select('张三', 20)
        [(3, '张三', 20)]
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').page(1, 10).select(name='张三', age=20)
        [(3, '张三', 20)]
        """
        return self.page_exec.select(self.sql, *args, **kwargs)

    def do_query(self, *args) -> List[dict]:
        """
        Execute select SQL and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').page(1, 10).do_query('张三', 20)
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        return self.page_exec.do_query(self.sql, *args)

    def do_select(self, *args) -> List[Tuple]:
        """
        Execute select SQL and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').page(1, 10).do_select('张三', 20)
        [(3, '张三', 20)]
        """
        return self.page_exec.do_select(self.sql, *args)


class ParamPageExec:

    def __init__(self, sql_page_exec: SqlPageExec, *args, **kwargs):
        self.sql_page_exec = sql_page_exec
        self.args = args
        self.kwargs = kwargs

    def query(self) -> List[dict]:
        """
        Execute select SQL and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').param('张三', 20).page(1, 10).query()
        [{'id': 3, 'name': '张三', 'age': 20}]
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').param(name='张三', age=20).page(1, 10).query()
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        return self.sql_page_exec.query(*self.args, **self.kwargs)

    def select(self) -> List[Tuple]:
        """
        Execute select SQL and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').param('张三', 20).page(1, 10).select()
        [(3, '张三', 20)]
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').param(name='张三', age=20).page(1, 10).select()
        [(3, '张三', 20)]
        """
        return self.sql_page_exec.select(*self.args, **self.kwargs)


class Param:

    def __init__(self, sql_exec, *args, **kwargs):
        self.sql_exec = sql_exec
        self.args = args
        self.kwargs = kwargs

    def execute(self) -> int:
        """
        Execute sql return effected rowcount

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').param('张三', 18).execute()
        1
        >>> db.sql('INSERT INTO person(name, age) VALUES(:name, :age)').param(name='张三', age=20).execute()
        1
        """
        return self.sql_exec.execute(*self.args, **self.kwargs)

    def save(self):
        """
        Insert data into table, return primary key.

        :param args:
        :return: Primary key

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').param('张三', 18).save()
        3
        >>> db.sql('INSERT INTO person(name, age) VALUES(:name, :age)').param(name='张三', age=20).save()
        3
        """
        return self.sql_exec.save(*self.args, **self.kwargs)

    def save_select_key(self, select_key: str):
        """
        Insert data into table, return primary key.

        :param select_key: sql for select primary key
        :param args:
        :return: Primary key

        Examples
        --------
        >>> import sqlexecx as db
        >>> select_key = 'SELECT LAST_INSERT_ID()'
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').param('张三', 20).save_select_key(select_key)
        3
        >>> db.sql('INSERT INTO person(name, age) VALUES(:name, :age)').param(name='张三', age=20).save_select_key(select_key)
        3
        """
        return self.sql_exec.save_select_key(select_key, *self.args, **self.kwargs)

    def get(self):
        """
        Execute select SQL and expected one int and only one int result, SQL contain 'limit'.

        MultiColumnsError: Expect only one column.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT count(1) FROM person WHERE name=? and age=? LIMIT 1').param('张三', 18).get()
        1
        >>> db.sql('SELECT count(1) FROM person WHERE name=:name and age=:age LIMIT 1').param(name='张三', age=20).get()
        1
        """
        return self.sql_exec.get(*self.args, **self.kwargs)

    def select(self) -> List[Tuple]:
        """
        Execute select SQL and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').param('张三', 18).select()
        [(3, '张三', 20)]
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').param(name='张三', age=20).select()
        [(3, '张三', 20)]
        """
        return self.sql_exec.select(*self.args, **self.kwargs)

    def select_one(self) -> Tuple:
        """
        Execute select SQL and return unique result(tuple), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=? LIMIT 1').param('张三', 18).select_one()
        (3, '张三', 20)
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age LIMIT 1').param(name='张三', age=20).select_one()
        (3, '张三', 20)
        """
        return self.sql_exec.select_one(*self.args, **self.kwargs)

    def query(self) -> List[dict]:
        """
        Execute select SQL and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').param('张三', 20).query()
        [{'id': 3, 'name': '张三', 'age': 20}]
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').param(name='张三', age=20).query()
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        return self.sql_exec.query(*self.args, **self.kwargs)

    def query_one(self) -> dict:
        """
        Execute select SQL and return unique result(dict), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').param('张三', 20).query_one()
        {'id': 3, 'name': '张三', 'age': 20}
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').param(name='张三', age=20).query_one()
        {'id': 3, 'name': '张三', 'age': 20}
        """
        return self.sql_exec.query_one(*self.args, **self.kwargs)

    def to_df(self):
        """
        Execute select SQL and return pandas DataFrame instance.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT * FROM person WHERE name=? and age=?').param('张三', 18).to_df()
        pd.DataFrame()
        """
        return self.sql_exec.load(*self.args, **self.kwargs).to_df()

    def to_csv(self, file_name: str, delimiter=',', header=True, encoding='utf-8'):
        """
        Execute select SQL and sava as a csv file.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT * FROM person WHERE name=? and age=?').param('张三', 18).to_csv('test.csv')
        """
        self.sql_exec.load(*self.args, **self.kwargs).to_csv(file_name, delimiter, header, encoding)

    def to_json(self, file_name: str, encoding='utf-8'):
        """
        Execute select SQL and sava as a json file.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').param('张三', 18).to_json('test.json')
        """
        self.sql_exec.load(*self.args, **self.kwargs).to_json(file_name, encoding)

    def page(self, page_num=1, page_size=10) -> ParamPageExec:
        """
        Execute select SQL and return ParamPageExec instance.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').param('张三', 18).page(1, 10)
        ParamPageExec()
        """
        return ParamPageExec(self.sql_exec.page(page_num, page_size), *self.args, **self.kwargs)


class SqlExec:

    def __init__(self, _exec, sql: str):
        self.exec = _exec
        self.sql = sql

    def execute(self, *args, **kwargs) -> int:
        """
        Execute sql return effected rowcount

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').execute('张三', 20)
        1
        >>> db.sql('INSERT INTO person(name, age) VALUES(:name, :age)').execute(name='张三', age=20)
        1
        """
        return self.exec.execute(self.sql, *args, **kwargs)

    def save(self, *args, **kwargs):
        """
        Insert data into table, return primary key.

        :param args:
        :return: Primary key

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').save('张三', 20)
        3
        >>> db.sql('INSERT INTO person(name, age) VALUES(:name, :age)').save(name='张三', age=20)
        3
        """
        return self.exec.save_sql(self.sql, *args, **kwargs)

    def save_select_key(self, select_key: str, *args, **kwargs):
        """
        Insert data into table, return primary key.

        :param select_key: sql for select primary key
        :param args:
        :return: Primary key

        Examples
        --------
        >>> import sqlexecx as db
        >>> select_key = 'SELECT LAST_INSERT_ID()'
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').save_select_key(select_key, '张三', 20)
        3
        >>> db.sql('INSERT INTO person(name, age) VALUES(:name, :age)').save_select_key(select_key, name='张三', age=20)
        3
        """
        return self.exec.save_sql_select_key(select_key, self.sql, *args, **kwargs)

    def get(self, *args, **kwargs):
        """
        Execute select SQL and expected one int and only one int result, SQL contain 'limit'.

        MultiColumnsError: Expect only one column.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT count(1) FROM person WHERE name=? and age=? LIMIT 1').get('张三', 20)
        1
        >>> db.sql('SELECT count(1) FROM person WHERE name=:name and age=:age LIMIT 1').get(name='张三', age=20)
        1
        """
        return self.exec.get(self.sql, *args, **kwargs)

    def select(self, *args, **kwargs) -> List[Tuple]:
        """
        Execute select SQL and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').select('张三', 20)
        [(3, '张三', 20)]
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').select(name='张三', age=20)
        [(3, '张三', 20)]
        """
        return self.exec.select(self.sql, *args, **kwargs)

    def select_one(self, *args, **kwargs) -> Tuple:
        """
        Execute select SQL and return unique result(tuple), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=? LIMIT 1').select_one('张三', 20)
        (3, '张三', 20)
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age LIMIT 1').select_one(name='张三', age=20)
        (3, '张三', 20)
        """
        return self.exec.select_one(self.sql, *args, **kwargs)

    def query(self, *args, **kwargs) -> List[dict]:
        """
        Execute select SQL and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').query('张三', 20)
        [{'id': 3, 'name': '张三', 'age': 20}]
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').query(name='张三', age=20)
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        return self.exec.query(self.sql, *args, **kwargs)

    def query_one(self, *args, **kwargs) -> dict:
        """
        Execute select SQL and return unique result(dict), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').query_one('张三', 20)
        {'id': 3, 'name': '张三', 'age': 20}
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').query_one(name='张三', age=20)
        {'id': 3, 'name': '张三', 'age': 20}
        """
        return self.exec.query_one(self.sql, *args, **kwargs)

    def do_execute(self, *args) -> int:
        """
        Execute sql return effected rowcount

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').do_execute('张三', 20)
        1
        """
        return self.exec.do_execute(None, self.sql, *args)

    def do_save_sql(self, select_key: str, *args):
        """
        Insert data into table, return primary key.

        :param select_key: sql for select primary key
        :param args:
        :return: Primary key

        Examples
        --------
        >>> import sqlexecx as db
        >>> select_key = 'SELECT LAST_INSERT_ID()'
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').do_save_sql(select_key, '张三', 20)
        3
        """
        return self.exec.do_save_sql(select_key, self.sql, *args)

    def do_get(self, *args):
        """
        Execute select SQL and expected one int and only one int result, SQL contain 'limit'.

        MultiColumnsError: Expect only one column.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT count(1) FROM person WHERE name=? and age=? LIMIT 1').do_get('张三', 20)
        1
        """
        return self.exec.do_get(self.sql, *args)

    def do_select(self, *args) -> List[Tuple]:
        """
        Execute select SQL and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=? LIMIT 1').do_select('张三', 20)
        (3, '张三', 20)
        """
        return self.exec.do_select(self.sql, *args)

    def do_select_one(self, *args) -> Tuple:
        """
        Execute select SQL and return unique result(tuple), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=? LIMIT 1').do_select_one('张三', 20)
        (3, '张三', 20)
        """
        return self.exec.do_select_one(self.sql, *args)

    def do_query(self, *args) -> List[dict]:
        """
        Execute select SQL and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').do_query('张三', 20)
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        return self.exec.do_query(self.sql, *args)

    def do_query_one(self, *args) -> dict:
        """
        execute select SQL and return unique result(dict), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').do_query_one('张三', 20)
        {'id': 3, 'name': '张三', 'age': 20}
        """
        return self.exec.do_query_one(self.sql, *args)

    def batch_execute(self, *args) -> int:
        """
        Batch execute sql return effected rowcount

        :param args: All number must have same size.
        :return: Effect rowcount

        Examples
        --------
        >>> import sqlexecx as db
        >>> args = [('张三', 20), ('李四', 28)]
        >>> db.sql('INSERT INTO person(name, age) VALUES(?, ?)').batch_execute(*args)
        2
        >>> args =  [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
        >>> db.sql('INSERT INTO person(name, age) VALUES(:name, :age)').batch_execute(sql, *args)
        2
        """
        return self.exec.batch_execute(self.sql, *args)

    def load(self, *args, **kwargs) -> Loader:
        """
        Execute select SQL and return a Loader instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').load('张三', 20)
        Lodder()
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').load(name='张三', age=20)
        Lodder()
        """
        return self.exec.load(self.sql, *args, **kwargs)

    def do_load(self, *args) -> Loader:
        """
        Execute select SQL and return a Loader instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').load('张三', 20)
        Lodder()
        """
        return self.exec.do_load(self.sql, *args)

    def param(self, *args, **kwargs) -> Param:
        """
        Get a Param instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').param('张三', 20)
        Param()
        >>> db.sql('SELECT id, name, age FROM person WHERE name=:name and age=:age').param(name='张三', age=20)
        Param()
        """
        return Param(self, *args, **kwargs)

    def page(self, page_num=1, page_size=10) -> SqlPageExec:
        """
        Get a SqlPageExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.sql('SELECT id, name, age FROM person WHERE name=? and age=?').page(1, 10)
        SqlPageExec()
        """
        return SqlPageExec(self.sql, PageExec(self.exec, page_num=page_num, page_size=page_size))


@lru_cache(maxsize=CACHE_SIZE)
def sql(sql_text: str) -> SqlExec:
    """
    Get a SqlExec instance

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.sql('SELECT id, name, age FROM person')
    """

    sql_text = sql_text.strip()
    assert sql_text, "Parameter 'sql' must not be none"
    return SqlExec(exec, sql_text)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

from functools import lru_cache
from typing import Tuple, List, Union
from sqlexecutorx import batch_execute as _batch_execute
from .loader import Loader
from . import exec, sql_support
from .dialect import Dialect, Engine
from .constant import LIMIT_1, SELECT_COUNT, CACHE_SIZE
from .table_support import get_table_select_sql, get_where_arg_limit, WhereBase
from .table_limit_exec import LimitExec, ColumnLimitExec, WhereLimitExec, ColumnWhereLimitExec
from .table_page_exec import TablePageExec, ColumnPageExec, WherePageExec, ColumnWherePageExec
from .table_order_by import OrderByExec, ColumnOrderByExec, WhereOrderByExec, ColumnWhereOrderByExec


class ColumnWhereExec:

    def __init__(self, where_exec, *columns):
        self._where_exec = where_exec
        self.columns = columns

    def get(self):
        """
        Select data from table and expected one int and only one int result, SQL contain 'limit'.
        MultiColumnsError: Expect only one column.

        sqlexecx.table('person').columns('name').where(id=1).get()
        """
        return self._where_exec.get(self.columns[0])

    def select(self) -> List[Tuple]:
        """
        Select data from table and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name='张三', age=20).select()
        [(3, '张三', 20)]
        """
        return self._where_exec.select(*self.columns)

    def select_one(self) -> Tuple:
        """
        Select data from table and return unique result(tuple), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name='张三', age=20).select_one()
        (3, '张三', 20)
        """
        return self._where_exec.select_one(*self.columns)

    def query(self) -> List[dict]:
        """
        Select data from table and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name='张三', age=20).query()
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        return self._where_exec.query(*self.columns)

    def query_one(self) -> dict:
        """
        Select from table and return unique result(dict), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name='张三', age=20).query_one()
        {'id': 3, 'name': '张三', 'age': 20}
        """
        return self._where_exec.query_one(*self.columns)

    def to_df(self):
        """
        Select from table and return pandas DataFrame instance.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name__eq='李四').to_df()
        """
        return self._where_exec.load(*self.columns).to_df()

    def to_csv(self, file_name: str, delimiter=',', header=True, encoding='utf-8'):
        """
        Select from table and sava as a csv file.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db..table('person').columns('id', 'name', 'age').where(name__eq='李四').to_csv('test.csv')
        """
        self._where_exec.load(*self.columns).to_csv(file_name, delimiter, header, encoding)

    def to_json(self, file_name: str, encoding='utf-8'):
        """
        Select from table and sava as a json file.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db..table('person').columns('id', 'name', 'age').where(name__eq='李四').to_json('test.json')
        """
        self._where_exec.load(*self.columns).to_json(file_name, encoding)

    def page(self, page_num=1, page_size=10, return_total=False) -> ColumnWherePageExec:
        """
        Get a ColumnWherePageExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name='张三', age=20).page(1, 10)
        ColumnWherePageExec()
        """
        return ColumnWherePageExec(self._where_exec.page(page_num, page_size, return_total), *self.columns)

    def limit(self, limit: Union[int, Tuple[int], List[int]] = 10) -> ColumnWhereLimitExec:
        """
        Get a ColumnWhereLimitExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name='张三', age=20).limit(10)
        ColumnWhereLimitExec()
        """
        return ColumnWhereLimitExec(self._where_exec.limit(limit), *self.columns)
    
    def order_by(self, order_by: str) -> ColumnWhereOrderByExec:
        """
        Get a ColumnWhereOrderByExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name='张三', age=20).order_by('id DESC, name ASC')
        ColumnWhereOrderByExec()
        """
        return ColumnWhereOrderByExec(self._where_exec.order_by(order_by), *self.columns)


class WhereExec(WhereBase):

    def __init__(self, _exec, table_name: str, order_by: str, **kwargs):
        super().__init__(_exec, table_name, order_by, **kwargs)

    def get(self, column: str):
        """
        Select data from table and expected one int and only one int result, SQL contain 'limit'.

        MultiColumnsError: Expect only one column.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(id=3).get('name')
        张三
        """
        sql, args = self.get_select_one_sql_args(column)
        return self.exec.do_get(sql, *args, LIMIT_1)

    def count(self) -> int:
        """
        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三').count()
        1
        """
        return self.get(SELECT_COUNT)

    def exists(self) -> bool:
        """
        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三').exists()
        True
        """
        return self.get(1) == 1

    def select_one(self, *columns) -> Tuple:
        """
        Select data from table and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三', age=20).select_one('id', 'name', 'age')
        (3, '张三', 20)
        """
        sql, args = self.get_select_one_sql_args(*columns)
        return self.exec.do_select_one(sql, *args, LIMIT_1)

    def query_one(self, *columns) -> dict:
        """
        Select from table and return unique result(dict), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三', age=20).query_one('id', 'name', 'age')
        {'id': 3, 'name': '张三', 'age': 20}
        """
        sql, args = self.get_select_one_sql_args(*columns)
        return self.exec.do_query_one(sql, *args, LIMIT_1)

    def delete(self) -> int:
        """
        Delete and return effect rowcount.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三').delete()
        1
        """
        where, args, _ = get_where_arg_limit(**self.where_condition)
        sql = 'DELETE FROM %s %s' % (Dialect.get_dialect_str(self.table), where)
        if Dialect.curr_engine() in (Engine.MYSQL, Engine.SQLITE):
            sql = '{} LIMIT ?'.format(sql)
            args = [*args, LIMIT_1]
        return self.exec.do_execute(sql, *args)

    def update(self, **kwargs) -> int:
        """
        Update and return effect rowcount.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三').update(name='李四', age=45)
        1
        """
        where, args, _ = get_where_arg_limit(**self.where_condition)
        update_cols, update_args = zip(*kwargs.items())
        args = [*update_args, *args]
        update_cols = ', '.join(['{} = ?'.format(Dialect.get_dialect_str(col)) for col in update_cols])
        sql = 'UPDATE {} SET {} {}'.format(Dialect.get_dialect_str(self.table), update_cols, where)
        if Dialect.curr_engine() in (Engine.MYSQL, Engine.SQLITE):
            sql = '{} LIMIT ?'.format(sql)
            args = [*args, LIMIT_1]
        return self.exec.do_execute(sql, *args)

    def columns(self, *columns) -> ColumnWhereExec:
        """
        Get a ColumnWhereExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三', age=20).columns('id', 'name', 'age')
        ColumnWhereExec()
        """
        return ColumnWhereExec(self, *columns)

    def page(self, page_num=1, page_size=10, return_total=False) -> WherePageExec:
        """
        Get a WherePageExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三', age=20).page(1, 10)
        WherePageExec()
        """
        return WherePageExec(self, page_num, page_size, return_total)

    def limit(self, limit: Union[int, Tuple[int], List[int]] = 10) -> WhereLimitExec:
        """
        Get a WhereLimitExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三', age=20).limit(10)
        WhereLimitExec()
        """
        return WhereLimitExec(self.exec, self.table, None, limit, **self.where_condition)
    
    def order_by(self, order_by: str) -> WhereOrderByExec:
        """
        Get a WhereOrderByExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').order_by('id DESC, name ASC')
        WhereOrderByExec()
        """
        return WhereOrderByExec(self.exec, self.table, order_by, **self.where_condition)
    
  
class ColumnExec:

    def __init__(self, table_exec, *columns):
        self.table_exec = table_exec
        self.columns = columns

    def insert(self, *args) -> int:
        """
        Insert data into table, return effect rowcount.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('name', 'age').insert('张三', 20)
        1
        """
        assert args, 'args expected but empty.'
        sql = sql_support.insert_sql(self.table_exec.table.strip(), self.columns)
        return self.table_exec.exec.execute(sql, *args)

    def batch_insert(self, *args) -> int:
        """
        Batch insert data into table and return effect rowcount

        Examples
        --------
        >>> import sqlexecx as db
        >>> args = [('张三', 20), ('李四', 28)]
        >>> db.table('person').columns('name', 'age').batch_insert(args)
        2
        >>> db.table('person').columns('name', 'age').batch_insert(*args)
        2
        """
        sql = sql_support.insert_sql(self.table_exec.table.strip(), self.columns)
        return _batch_execute(sql, *args)

    def get(self):
        """
        Select data from table and expected one int and only one int result, SQL contain 'limit'.

        MultiColumnsError: Expect only one column.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('count(1)').get()
        3
        """
        return self.table_exec.get(*self.columns)

    def select(self) -> List[Tuple]:
        """
        Select data from table and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').select()
        [(3, '张三', 20)]
        """
        return self.table_exec.select(*self.columns)

    def select_one(self) -> Tuple:
        """
        Select data from table and return unique result(tuple), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').select_one()
        (3, '张三', 20)
        """
        return self.table_exec.select_one(*self.columns)

    def query(self) -> List[dict]:
        """
        Select data from table and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').query()
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        return self.table_exec.query(*self.columns)

    def query_one(self) -> dict:
        """
        Select from table and return unique result(dict), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').query_one()
        {'id': 3, 'name': '张三', 'age': 20}
        """
        return self.table_exec.query_one(*self.columns)

    def to_df(self):
        """
        Select from table and return pandas DataFrame instance.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').to_df()
        """
        return self.table_exec.load(*self.columns).to_df()

    def to_csv(self, file_name: str, delimiter=',', header=True, encoding='utf-8'):
        """
        Select from table and sava as a csv file.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db..table('person').columns('id', 'name', 'age').to_csv('test.csv')
        """
        self.table_exec.load(*self.columns).to_csv(file_name, delimiter, header, encoding)

    def to_json(self, file_name: str, encoding='utf-8'):
        """
        Select from table and sava as a json file.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db..table('person').columns('id', 'name', 'age').to_json('test.json')
        """
        self.table_exec.load(*self.columns).to_json(file_name, encoding)

    def where(self, **kwargs) -> ColumnWhereExec:
        """
        Get a ColumnWhereExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').where(name='张三', age=20)
        ColumnWhereExec()
        """
        return ColumnWhereExec(self.table_exec.where(**kwargs), *self.columns)

    def page(self, page_num=1, page_size=10, return_total=False) -> ColumnPageExec:
        """
        Get a ColumnPageExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').page(1, 10)
        ColumnPageExec()
        """
        return ColumnPageExec(self.table_exec.page(page_num, page_size, return_total), *self.columns)

    def limit(self, limit: Union[int, Tuple[int], List[int]] = 10) -> ColumnLimitExec:
        """
        Get a ColumnLimitExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').limit(10)
        ColumnLimitExec()
        """
        return ColumnLimitExec(self.table_exec.limit(limit), *self.columns)
    
    def order_by(self, order_by: str) -> ColumnOrderByExec:
        """
        Get a ColumnOrderByExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age').order_by('id DESC, name ASC')
        ColumnOrderByExec()
        """
        return ColumnOrderByExec(self.table_exec.order_by(order_by), *self.columns)


class TableExec:

    def __init__(self, _exec, table_name):
        self.exec = _exec
        self.table = table_name

    def insert(self, **kwargs) -> int:
        """
        Insert data into table, return effect rowcount.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').insert(name='张三', age=20)
        1
        """
        return self.exec.insert(self.table, **kwargs)

    def save(self, **kwargs):
        """
        Insert data into table, return primary key.

        :return: Primary key

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').save(name='张三', age=20)
        3
        """
        return self.exec.save(self.table, **kwargs)

    def save_select_key(self, select_key: str, **kwargs):
        """
        Insert data into table, return primary key.

        :param select_key: sql for select primary key
        :return: Primary key

        Examples
        --------
        >>> import sqlexecx as db
        >>> select_key = 'SELECT LAST_INSERT_ID()'
        >>> db.table('person').save_select_key(select_key, name='张三', age=20)
        3
        """
        return self.exec.save_select_key(select_key, self.table, **kwargs)

    def batch_insert(self, *args) -> int:
        """
        Batch insert data into table and return effect rowcount

        Examples
        --------
        >>> import sqlexecx as db
        >>> args = [{'name': '张三', 'age': 20}, {'name': '李四', 'age': 28}]
        >>> db.table('person').batch_insert(*args)
        2
        """
        return self.exec.batch_insert(self.table, *args)

    def get(self, column: str):
        """
        Select data from table and expected one int and only one int result, SQL contain 'limit'.

        MultiColumnsError: Expect only one column.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').get('count(1)')
        3
        """
        sql = get_table_select_sql(self.table, None, None, LIMIT_1, column)
        return self.exec.do_get(sql, LIMIT_1)

    def count(self) -> int:
        """
        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').count()
        3
        """
        return self.get(SELECT_COUNT)

    def select(self, *columns) -> List[Tuple]:
        """
        Select data from table and return list results(tuple).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').select('id', 'name', 'age')
        [(3, '张三', 20)]
        """
        sql = get_table_select_sql(self.table, None, None, 0, *columns)
        return self.exec.do_select(sql)

    def select_one(self, *columns) -> Tuple:
        """
        Select data from table and return unique result(tuple), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').do_select_one('id', 'name', 'age')
        (3, '张三', 20)
        """
        sql = get_table_select_sql(self.table, None, None, LIMIT_1, *columns)
        return self.exec.do_select_one(sql, LIMIT_1)

    def query(self, *columns) -> List[dict]:
        """
        Select data from table and return list results(dict).

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').query('id', 'name', 'age')
        [{'id': 3, 'name': '张三', 'age': 20}]
        """
        sql = get_table_select_sql(self.table, None, None, 0, *columns)
        return self.exec.do_query(sql)

    def query_one(self, *columns) -> dict:
        """
        Select data from table and return unique result(dict), SQL contain 'limit'.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').query_one('id', 'name', 'age')
        {'id': 3, 'name': '张三', 'age': 20}
        """
        sql = get_table_select_sql(self.table, None, None, LIMIT_1, *columns)
        return self.exec.do_query_one(sql, LIMIT_1)

    def load(self, *columns) -> Loader:
        """
        Select from table and return a Loader instance

        :return: Loader

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').load('id', 'name', 'age')
        Lodder()
        """
        sql = get_table_select_sql(self.table, None, None, 0, *columns)
        return self.exec.do_load(sql)

    def insert_from_df(self, df, columns: Tuple[str] = None):
        """
        Insert data into table from pandas DataFrame and return effected rowcount.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').insert_from_df(df)
        20
        """
        return self.exec.insert_from_df(df, self.table, columns)

    def insert_from_csv(self, file_name: str, delimiter=',', header=True, columns: Tuple[str] = None, encoding='utf-8'):
        """
        Insert data into table from a csv file and return effected rowcount.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').insert_from_csv('test.csv')
        20
        """
        return self.exec.insert_from_csv(file_name, self.table, delimiter, header, columns, encoding=encoding)

    def insert_from_json(self, file_name: str, encoding='utf-8'):
        """
        Insert data into table from a json file and return effected rowcount.

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').insert_from_json('test.json')
        20
        """
        return self.exec.insert_from_json(file_name, self.table, encoding=encoding)

    def truncate(self) -> int:
        """
        Truncate table and return effected rowcount

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').truncate()
        1
        """
        return self.exec.truncate(self.table)

    def drop(self) -> int:
        """
        Drop table and return effected rowcount

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').drop()
        1
        """
        return self.exec.drop(self.table)

    def where(self, **kwargs) -> WhereExec:
        """
        Get a WhereExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').where(name='张三', age=20)
        WhereExec()
        """
        return WhereExec(self.exec, self.table, None, **kwargs)

    def columns(self, *columns) -> ColumnExec:
        """
        Get a ColumnExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').columns('id', 'name', 'age')
        ColumnExec()
        """
        return ColumnExec(self, *columns)

    def page(self, page_num=1, page_size=10, return_total=False) -> TablePageExec:
        """
        Get a TablePageExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').page(1, 10)
        TablePageExec()
        """
        return TablePageExec(self, page_num, page_size, return_total)

    def limit(self, limit: Union[int, Tuple[int], List[int]] = 10) -> LimitExec:
        """
        Get a LimitExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').limit(10)
        LimitExec()
        """
        return LimitExec(self.exec, self.table, None, limit)
    
    def order_by(self, order_by: str) -> OrderByExec:
        """
        Get a OrderByExec instance

        Examples
        --------
        >>> import sqlexecx as db
        >>> db.table('person').order_by('id DESC, name ASC')
        OrderByExec()
        """
        return OrderByExec(self.exec, self.table, order_by)


@lru_cache(maxsize=CACHE_SIZE)
def table(table_name: str) -> TableExec:
    """
    Get a TableExec instance

    Examples
    --------
    >>> import sqlexecx as db
    >>> db.table('person')
    """
    table_name = table_name.strip()
    assert table_name, "Parameter 'table' must not be none"
    return TableExec(exec, table_name)



from sqlexecutorx.log_support import logger, do_sql_log, do_save_log, batch_sql_log, db_ctx_log


def insert_log(function: str, table: str, **kwargs):
    logger.debug("Exec func 'sqlexecx.%s' \n\t Table: '%s', kwargs: %s" % (function, table, kwargs))


def save_log(function: str, select_key: str, table: str, **kwargs):
    logger.debug("Exec func 'sqlexecx.%s', 'select_key': %s \n\t Table: '%s', kwargs: %s" % (function, select_key, table, kwargs))


def sql_log(function: str, sql: str, *args, **kwargs):
    logger.debug("Exec func '%s' \n\tsql: %s \n\targs: %s \n\tkwargs: %s" % (function, sql.strip(), args, kwargs))


def do_sql_log(function: str, sql: str, *args):
    logger.debug("Exec func '%s' \n\t sql: %s \n\t args: %s" % (function, sql, args))

#
# def db_ctx_log(action, connection):
#     logger.debug("%s connection <%s>..." % (action, hex(id(connection))))

